module StartHer
  module Subscriber
    include StartHer::RedisClient
    include StartHer::Utils

    attr_reader :channels, :options, :subscriber_heartbeat_block, :subscriber_on_psubscribe_block

    DEFAULT_OPTS = { channels: ['*'], heartbeat: Heartbeat::CHANNELS }
    DEFAULT_ERROR_BLOCK = lambda do |e|
      StartHer.logger.error e
    end

    def self.included(base)
      base.extend(ClassMethods)
    end

    module ClassMethods
      attr_accessor :subscriber_options_hash, :subscriber_error_block, :subscriber_heartbeat_block,
                    :subscriber_on_psubscribe_block

      def run!
        new(subscriber_options_hash[:channels]).run!
      end

      def subscriber_heartbeat(&block)
        self.subscriber_heartbeat_block = block
      end

      def subscriber_on_psubscribe(&block)
        self.subscriber_on_psubscribe_block = block
      end

      def subscriber_error(&block)
        self.subscriber_error_block = block
      end

      def subscriber_options(opts = {})
        self.subscriber_options_hash = DEFAULT_OPTS.merge(opts)
      end
    end

    def initialize(channels)
      @channels = channels
      @options = self.class.subscriber_options_hash || DEFAULT_OPTS
      @subscriber_heartbeat_block = self.class.subscriber_heartbeat_block || ->(_response) {}
      @subscriber_on_psubscribe_block = self.class.subscriber_on_psubscribe_block || ->(_chan) {}
    end

    def run!
      psubscribe(channels) do |channel, message|
        begin
          process_message(channel, message)
        rescue => e
          if block_given?
            subscriber_error_block.call(e)
          else
            DEFAULT_ERROR_BLOCK.call(e)
          end
        end
      end
    end

    # rubocop:disable Metrics/AbcSize,Metrics/MethodLength
    def psubscribe(channels, &block)
      exponential_backoff({}, ::Redis::BaseConnectionError) do
        client.psubscribe(options[:heartbeat][:in], *channels) do |on|
          on.psubscribe do |channel, _|
            StartHer.logger.info "PSuscribe on #{channel}"

            resync_messages(client, channel, &block)
            subscriber_on_psubscribe_block.call(unamespace(channel))
          end

          on.pmessage do |_pattern, channel, message|
            chan = unamespace(channel)
            msg = MessagePack.unpack(message)

            if chan == options[:heartbeat][:in]
              Heartbeat.call(msg, service_klass,
                heartbeat: options[:heartbeat], &subscriber_heartbeat_block)
            else
              block.call(chan, msg)
            end
          end
        end
      end
    rescue => e
      StartHer.logger.error e
      raise e
    end
    # rubocop:enable Metrics/AbcSize,Metrics/MethodLength

    def resync_messages(client, channel, &block)
      StartHer.logger.info "Re-synchronize message from #{channel}"

      # Retrieve all backlogs for a given channel
      client.keys(unamespace(channel) + '*').each do |backlog|
        # Retrieve all messages for a given backlog
        client.lrange(backlog, 0, -1).each do |message|
          block.call(backlog.split(':').first, MessagePack.unpack(message))
        end
      end
    end

    private

    def unamespace(channel)
      channel.split(':').last
    end
  end
end
